#!/bin/bash
# Regression testing script for MicroC
# Step through a list of files
#  Compile, run, and check the output of each expected-to-work test
#  Compile and check the error of each expected-to-fail test

# Path to the LLVM interpreter
LLI="lli"
#LLI="/usr/local/opt/llvm/bin/lli"

# Path to the LLVM compiler
LLC="llc"

# Path to the C compiler
CC="cc"

# Path to the microc compiler.  Usually "./microc.native"
# Try "_build/microc.native" if ocamlbuild was unable to create a symbolic link.
GMM="./gmm.native"
#MICROC="_build/microc.native"

globallog=testall.log
rm -f $globallog
error=0
globalerror=0
src=./src/

SignalError() {
    if [ $error -eq 0 ] ; then
	echo "FAILED"
	error=1
    fi
    echo "  $1"
}

# Run <args>
# Report the command, run it, and report any errors
Run() {
    echo $* 1>&2
    eval $* || {
	SignalError "$1 failed on $*"
	return 1
    }
}

Check() {
    error=0
    basename=`echo $1 | sed 's/.*\\///
                             s/.gmm//'`
    # reffile=`echo $1 | sed 's/.mc$//'`
    basedir="`echo $1 | sed 's/\/[^\/]*$//'`/."

    echo -n "$basename..."

    echo 1>&2
    echo "###### Testing $basename" 1>&2

    generatedfiles=""

    generatedfiles="$generatedfiles ${basename}.ll ${basename}.s ${basename}.exe ${basename}.out" &&
    Run "$GMM" "$1" ">" "${basename}.ll" &&
    Run "$LLC" "-relocation-model=pic" "${basename}.ll" ">" "${basename}.s" &&
    Run "$CC" "-o" "${basename}.exe" "${basename}.s" "-pthread" "${src}builtin.o" &&
    Run "./${basename}.exe"
    # Run "./${basename}.exe" > "${basename}.out" # &&
    # Compare ${basename}.out ${reffile}.out ${basename}.diff

    # # Report the status and clean up the generated files

    # if [ $error -eq 0 ] ; then
	# if [ $keep -eq 0 ] ; then
	#     rm -f $generatedfiles
	# fi
	# echo "OK"
	# echo "###### SUCCESS" 1>&2
    # else
	# echo "###### FAILED" 1>&2
	# globalerror=$error
    # fi
}

Check $1
# rm *.ll *.s *.exe

