/*
 * builtin functions and functionalites for Go--
 */
#include <stdio.h>
#include <pthread.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdbool.h>
#include <string.h>
#include <semaphore.h>
#include <sys/stat.h>   
#include <pthread.h>
#include <errno.h>

#define N_THREADS 16

//code adapted from cs4118 fall 2020

int init = 0;

static pthread_t thread_pool[N_THREADS];

struct node {
    void *args; // Payload, in our case a new gocall
    struct node *next; // Next call on the list
};

pthread_mutex_t mp  = PTHREAD_MUTEX_INITIALIZER;

struct queue {
    pthread_mutex_t mutex; // A mutex used to protect the queue itself
    pthread_cond_t cond;   // A condition variable for threads to sleep on
    struct node *first; // First message in the queue
    struct node *last;  // Last message in the queue
    unsigned int length;   // Number of elements on the queue
};

static struct queue thread_queue;

static void die(const char *message)
{
    perror(message);
    exit(1);
}


static void queue_init(struct queue *q)
{
    memset(q, 0, sizeof(*q));

    if (pthread_cond_init(&q->cond, NULL) != 0)
        die("Cannot initialize queue condition variable");

    if (pthread_mutex_init(&q->mutex, NULL) != 0) {
        pthread_cond_destroy(&q->cond);
        die("Cannot initialize queue mutex");
    }
}


static void queue_destroy(struct queue *q)
{
    struct node *tmp;
    // First of all delete all elements currently on the queue
    pthread_mutex_lock(&q->mutex);
    while(q->first) {
        tmp = q->first;
        q->first = q->first->next;
        free(tmp);
    }
    q->last = NULL;
    q->length = 0;
    pthread_mutex_unlock(&q->mutex);
    // After that destroy the mutex and the condition variable
    pthread_mutex_destroy(&q->mutex);
    pthread_cond_destroy(&q->cond);
}



static void queue_put(struct queue *q, void *args)
{
    struct node *msg;

    // Create a new message object, initialized to all zeroes and store the
    // socket file descriptor in its sock attribute
    msg = (struct node *)calloc(1, sizeof(*msg));
    if (msg == NULL)
        die("Out of memory");
    msg->args = args;

    // Append the new message to the queue
    pthread_mutex_lock(&q->mutex);
    if (q->last == NULL) {
        q->last = msg;
        q->first = msg;
    } else {
        q->last->next = msg;
        q->last = msg;
    }
    // If the queue was previously empty, wakep up any threads that might be
    // sleeping, waiting for the queue to contain data
    if (q->length == 0)
        pthread_cond_broadcast(&q->cond);
    q->length++;
    pthread_mutex_unlock(&q->mutex);
}


void *queue_get(struct queue *q)
{
    void *args;
    struct node *msg;

    pthread_mutex_lock(&q->mutex);

    /* Since all the sleeping threads content for the data on the queue, we
     * need to check if the queue is non-empty after we woke up. If yes we
     * won, if not we go back to sleep on the condition variable.
     * pthread_cond_wait unlocks the give mutex and suspends the thread.
     */
    while (q->first == NULL) {
        pthread_cond_wait(&q->cond, &q->mutex);
    }

    // We won this round, remove the data from the queue and return it.
    msg = q->first;
    q->first = q->first->next;
    q->length--;

    if (q->first == NULL) {
        q->last = NULL;
        q->length = 0;
    }

    args = msg->args;
    free(msg);
    pthread_mutex_unlock(&q->mutex);
    return args;
}




/* ------------------------- string builtins -------------------------------*/


char *string_cat(char *a,char *b)
{
  int len = strlen(a) + strlen(b) + 1;
  char* result = malloc(len);
  strcat(result, a);
  strcat(result, b);
  return result;
}

int string_cmp(char *a,char *b)
{
  int result = strcmp(a,b);
  return result;
}

/* funtion to print double in binary,for debugs only*/
void printbin(double x)
{
    union {
        double x;
        char c[sizeof(double)];
    } u;
    u.x = x;

    for (unsigned ofs = 0; ofs < sizeof(double); ofs++) {
        for(int i = 7; i >= 0; i--) {
            printf(((1 << i) & u.c[ofs]) ? "1" : "0");
        }
        printf(" ");
    }
}

/* ------------------------- gofunction builtins -------------------------------*/
/* function to execute the gofunction and manange memory */
void* gothread(void* data)
{
    for(;;){
        void *params = queue_get(&thread_queue);
        void **ptr_f =(void **) params;
        
        void *fptr = *ptr_f; // first one is func ptr
        // ptr to parameters
        int *ptr = ((int *)params)+sizeof(void *)/sizeof(int);
        int size = *(ptr++); // get number of parameters
        int sum;
        int typs[size];
        double *dptr;
        char **sptr;
        
        for(int i = 0; i < size; i++){
            typs[i] = *(ptr++);
        }
        
        
        if (size > 3) {
            printf("%s\n","argument exceeds maximum -> 3");
        }
        if (size == 0){
            void (*f)() = (void (*)())fptr;
            f(); 
        }
        if (size == 1){
            if(typs[0] == 1){
                void(*f)(int) = (void (*) (int))fptr;
                f(*ptr);
            } else if (typs[0] == 2){
                void(*f)(double) = (void (*) (double))fptr;
                f(*((double *)ptr));
            } else if (typs[0] == 3){
                void(*f)(char *) = (void (*) (char *))fptr;
                f(*((char **)ptr));
            }
        }
        if (size == 2){
            if(typs[0]==1){
                int a1 = *ptr;
                ptr++;
                if(typs[1]==1){
                    int a2 = *ptr;
                    void(*f)(int,int) = (void (*) (int,int))fptr;
                    f(a1,a2);
                }else if (typs[1]==2){
                    dptr = (double *)ptr;
                    double a2 = *dptr;
                    void(*f)(int,double) = (void (*) (int,double))fptr;
                    f(a1,a2);
                }else if (typs[1]==3){
                    sptr = (char **)ptr;
                    char *a2 = *sptr;
                    void(*f)(int,char *) = (void (*) (int,char *))fptr;
                    f(a1,a2);
                }
            }else if(typs[0]==2){
                dptr =(double *)ptr;
                double a1 = *dptr;
                dptr++;
                if(typs[1]==1){
                    ptr = (int *)dptr;
                    int a2 = *ptr;
                    void(*f)(double,int) = (void (*) (double,int))fptr;
                    f(a1,a2);
                }else if (typs[1]==2){
                    double a2 = *dptr;
                    void(*f)(double,double) = (void (*) (double,double))fptr;
                    f(a1,a2);
                }else if (typs[1]==3){
                    sptr = (char **)dptr;
                    char *a2 = *sptr;
                    void(*f)(double,char *) = (void (*) (double,char *))fptr;
                    f(a1,a2);
                }

            }else if(typs[0]==3){
                sptr =(char **)ptr;
                char  *a1 = *sptr;
                sptr++;
                if(typs[1]==1){
                    ptr = (int *)sptr;
                    int a2 = *ptr;
                    void(*f)(char *,int) = (void (*) (char *,int))fptr;
                    f(a1,a2);
                }else if (typs[1]==2){
                    dptr = (double *)sptr;
                    double a2 = *dptr;
                    void(*f)(char *,double) = (void (*) (char *,double))fptr;
                    f(a1,a2);
                }else if (typs[1]==3){
                    char *a2 = *sptr;
                    void(*f)(char *,char *) = (void (*) (char *,char *))fptr;
                    f(a1,a2);
                }
                
            }
        }
        if (size == 3){
            if(typs[0] == 1){
                int a1 = *ptr;
                ptr++;
                if(typs[1] == 1){
                    int a2= *ptr;
                    ptr++;
                    if(typs[2] == 1){
                        int a3 = *ptr;
                        void(*f)(int,int,int) = (void (*) (int,int,int))fptr;
                        f(a1,a2,a3);
                    }else if (typs[2] == 2){
                        dptr = (double *)ptr;
                        double a3 = *dptr;
                        void(*f)(int,int,double) = (void (*) (int,int,double))fptr;
                        f(a1,a2,a3);
                    }else if (typs[2] == 3){
                        sptr = (char **)ptr;
                        char *a3 = *sptr;
                        void(*f)(int,int,char *) = (void (*) (int,int,char *))fptr;
                        f(a1,a2,a3);
                    }
                }else if(typs[1] == 2){
                    dptr =(double *)ptr;
                    double a2= *dptr;
                    dptr++;
                    if(typs[2] == 1){
                        ptr =(int *)dptr;
                        int a3 = *ptr;
                        void(*f)(int,double,int) = (void (*) (int,double,int))fptr;
                        f(a1,a2,a3);
                    }else if (typs[2] == 2){
                        double a3 = *dptr;
                        void(*f)(int,double,double) = (void (*) (int,double,double))fptr;
                        f(a1,a2,a3);
                    }else if (typs[2] == 3){
                        sptr = (char **)dptr;
                        char *a3 = *sptr;
                        void(*f)(int,double,char *) = (void (*) (int,double,char *))fptr;
                        f(a1,a2,a3);
                    }
                }else if(typs[1] == 3){
                    sptr =(char **)ptr;
                    char *a2= *sptr;
                    sptr ++;
                    if(typs[2] == 1){
                        ptr =(int *)sptr;
                        int a3 = *ptr;
                        void(*f)(int,char *,int) = (void (*) (int,char *,int))fptr;
                        f(a1,a2,a3);
                    }else if (typs[2] == 2){
                        dptr = (double *)sptr;
                        double a3 = *dptr;
                        void(*f)(int,char *,double) = (void (*) (int,char *,double))fptr;
                        f(a1,a2,a3);
                    }else if (typs[2] == 3){
                        sptr = (char **)sptr;
                        char *a3 = *(char **)sptr;
                        void(*f)(int,char *,char *) = (void (*) (int,char *,char *))fptr;
                        f(a1,a2,a3);
                    }
                }
            } else if(typs[0] == 2){
                dptr = (double *)ptr;
                double a1 = *dptr;
                dptr++;
                if(typs[1] == 1){
                    ptr = (int *)dptr;
                    int a2= *ptr;
                    ptr++;
                    if(typs[2] == 1){
                        int a3 = *ptr;
                        void(*f)(double,int,int) = (void (*) (double,int,int))fptr;
                        f(a1,a2,a3);
                    }else if (typs[2] == 2){
                        dptr = (double *)ptr;
                        double a3 = *dptr;
                        void(*f)(double,int,double) = (void (*) (double,int,double))fptr;
                        f(a1,a2,a3);
                    }else if (typs[2] == 3){
                        sptr = (char **)ptr;
                        char *a3 = *sptr;
                        void(*f)(double,int,char *) = (void (*) (double,int,char *))fptr;
                        f(a1,a2,a3);
                    }
                }else if(typs[1] == 2){
                    double a2= *dptr;
                    dptr++;
                    if(typs[2] == 1){
                        ptr =(int *)dptr;
                        int a3 = *ptr;
                        void(*f)(double,double,int) = (void (*) (double,double,int))fptr;
                        f(a1,a2,a3);
                    }else if (typs[2] == 2){
                        double a3 = *dptr;
                        void(*f)(double,double,double) = (void (*) (double,double,double))fptr;
                        f(a1,a2,a3);
                    }else if (typs[2] == 3){
                        sptr = (char **)dptr;
                        char *a3 = *sptr;
                        void(*f)(double,double,char *) = (void (*) (double,double,char *))fptr;
                        f(a1,a2,a3);
                    }
                }else if (typs[1] == 3){
                    sptr = (char **)dptr;
                    char *a2 = *sptr;
                    sptr++;
                    if(typs[2] == 1){
                        ptr =(int *)sptr;
                        int a3 = *ptr;
                        void(*f)(double,char *,int) = (void (*) (double,char *,int))fptr;
                        f(a1,a2,a3);
                    }else if (typs[2] == 2){
                        dptr =(double *)sptr;
                        double a3 = *dptr;
                        void(*f)(double,char *,double) = (void (*) (double,char *,double))fptr;
                        f(a1,a2,a3);
                    }else if (typs[2] == 3){
                        char *a3 = *sptr;
                        void(*f)(double,char *,char *) = (void (*) (double,char *,char *))fptr;
                        f(a1,a2,a3);
                    }
                    
                }
            } else if (typs[0] == 3){
                sptr = (char **)ptr;
                char *a1 = *sptr;
                sptr++;
                if(typs[1] == 1){
                    ptr = (int *)sptr;
                    int a2= *ptr;
                    ptr++;
                    if(typs[2] == 1){
                        int a3 = *ptr;
                        void(*f)(char *,int,int) = (void (*) (char *,int,int))fptr;
                        f(a1,a2,a3);
                    }else if (typs[2] == 2){
                        dptr = (double *)ptr;
                        double a3 = *dptr;
                        void(*f)(char *,int,double) = (void (*) (char *,int,double))fptr;
                        f(a1,a2,a3);
                    }else if (typs[2] == 3){
                        sptr = (char **)ptr;
                        char *a3 = *sptr;
                        void(*f)(char *,int,char *) = (void (*) (char *,int,char *))fptr;
                        f(a1,a2,a3);
                    }
                }else if(typs[1] == 2){
                    dptr = (double *)sptr;
                    double a2= *dptr;
                    dptr++;
                    if(typs[2] == 1){
                        ptr =(int *)dptr;
                        int a3 = *ptr;
                        void(*f)(char *,double,int) = (void (*) (char *,double,int))fptr;
                        f(a1,a2,a3);
                    }else if (typs[2] == 2){
                        double a3 = *dptr;
                        void(*f)(char *,double,double) = (void (*) (char *,double,double))fptr;
                        f(a1,a2,a3);
                    }else if (typs[2] == 3){
                        sptr = (char **)dptr;
                        char *a3 = *sptr;
                        void(*f)(char *,double,char *) = (void (*) (char *,double,char *))fptr;
                        f(a1,a2,a3);
                    }
                }else if (typs[1] == 3){
                    char *a2 = *sptr;
                    sptr++;
                    if(typs[2] == 1){
                        ptr =(int *)sptr;
                        int a3 = *ptr;
                        void(*f)(char *,char *,int) = (void (*) (char *,char *,int))fptr;
                        f(a1,a2,a3);
                    }else if (typs[2] == 2){
                        dptr =(double *)sptr;
                        double a3 = *dptr;
                        void(*f)(char *,char *,double) = (void (*) (char *,char *,double))fptr;
                        f(a1,a2,a3);
                    }else if (typs[2] == 3){
                        char *a3 = *sptr;
                        void(*f)(char *,char *,char *) = (void (*) (char *,char *,char *))fptr;
                        f(a1,a2,a3);
                    }
                }
            }
        }
        free((char *)params);
    }
}



void gocall(void* func, void * pre,void * ints, void * doubles,void *sandsts)
{
	pthread_t tid;
    
    pthread_mutex_lock(&mp);
    // if has already been init'd
    if(init !=0 ){
        goto proceed;
    }
	/* get the underlying structure of argument
	 * and wrap it into memory on the c heap
	 */
    init = 1;
    queue_init(&thread_queue);
    for(int i = 0; i < N_THREADS; i++) {
        pthread_create(&thread_pool[i], NULL, gothread, NULL);
    }
proceed:
    pthread_mutex_unlock(&mp);
	int *ptr =(int *) pre; //cast pre to int*
	int size = *(ptr++); // get the number of parameters
	int typs[size];
	// sorted pointers to int and double values
	int *intptr = (int *) ints;
	double * dptr = (double *) doubles; 
	char **pptr = (char **) sandsts;
	// calculate the space needed to store all params on heap
	int len = 0;
	for (int i = 0; i < size; i++){
		typs[i] = *(ptr+i);
		if(typs[i] == 1){ //int
			len+=4;
		}else if (typs[i] == 2){ //float
			len+=8;
		}else if (typs[i] == 3){
			len+=8;
		}
    }

	char *pt = malloc(sizeof(void *)+(size+1)*sizeof(int)+len);
	char *spt = memcpy(pt,&func,sizeof(void *));
	spt += sizeof(void *);
	spt = memcpy(spt,(void *)pre,(size+1)*sizeof(int));
	spt += (size+1)*sizeof(int);
	for (int i = 0; i < size; i++){
		typs[i] = *(ptr+i);
	}

	for (int i =0; i < size ;i ++) {
        if(typs[i] == 1){
            spt = memcpy(spt,intptr,sizeof(int));
			spt += sizeof(int);
			intptr++;
        }else if (typs[i] == 2){
            spt = memcpy(spt,dptr,sizeof(double));
			spt += sizeof(double);
            dptr++;	
        }else if (typs[i] == 3){
			spt = memcpy(spt,pptr,sizeof(char *));
            spt += sizeof(char *);
			pptr++;
		}
    }


   queue_put(&thread_queue, pt);
}


/* -------------------------- channel builtins ----------------------------*/

int *wrapAround(int *idx, int* size){
	(*idx)++;
	if (*idx == *size)
		*idx = 0;
	return idx;
}

/*
@a: pointer to array storing values
@b: pointer to int array for book-keeping [cap, start, end, type_id]
@c: pointer to pthread_mutex_t array [data_lk]
@d: pointer to sem_t array [sem_empty, sem_full]
@val: pointer to value to be enqueued
*/
void enqueue_f(void* a, void* b, void* c, void* d, void* val)
{
	// fixed type fields of a channel
	int* bk = (int*)b;
	int* cap = bk;
	int* start = bk + 1;
	int* end = bk + 2;
	int* type_id = bk + 3;
	pthread_mutex_t* lks = (pthread_mutex_t* )c;
	sem_t* sems = (sem_t* )d;
	sem_wait(sems+1);
	pthread_mutex_lock(lks);
	// variable type fields of a channel
	switch(*type_id)
	{
		case 0: //int
			*((int* )a + *end) = *(int* )val;
			break;
		case 1: //float or double
			*((double* )a + *end) = *(double* )val;
			break;
		case 2: //bool
			*((bool* )a + *end) = *(bool* )val;
			break;
        case 3: //string
            *((char**)a + *end) = *(char**)val;
            break;
        case 4: //struct
            *((void**)a + *end) = *(void**)val;
            break;
		default:
			printf("type_id %d is not supported by enqueue_f\n", *type_id);
	}
	wrapAround(end, cap);
	pthread_mutex_unlock(lks);
	sem_post(sems);
}


/*
@a: pointer to array storing values
@b: pointer to int array for book-keeping [cap, start, end, type_id]
@c: pointer to pthread_mutex_t array [data_lk]
@d: pointer to sem_t array [sem_empty, sem_full]
@val: pointer to space to dequeue into
*/
void dequeue_f(void* a, void* b, void* c, void* d, void* val)
{

	int* bk = (int*)b;
	int* cap = bk;
	int* start = bk + 1;
	int* end = bk + 2;
	int* type_id = bk + 3;
	pthread_mutex_t* lks = (pthread_mutex_t* )c;
	sem_t* sems = (sem_t* )d;

	sem_wait(sems);
	pthread_mutex_lock(lks);
	switch(*type_id)
	{
		case 0: //int
			*(int*) val = *((int*)a + *start);
			break;
		case 1: //float or double
			*(double*) val = *((double*)a + *start);
			break;
		case 2: //bool
			*(bool*) val = *((bool*)a + *start);
			break;
        case 3: //string
            *(char**) val = *((char**)a + *start);
            break;
        case 4: //struct
            *(void**)val = *((void**)a + *start);
            break;
		default:
			printf("type_id %d is not supported by dequeue_f\n", *type_id);
	}
	
	wrapAround(start, cap);
	pthread_mutex_unlock(lks);
	sem_post(sems+1);
}


